#define ledR D6
#define ledG D5
#define ledB D7
#define pinTemp 3

#include <Arduino.h>
#include <ESP8266WiFi.h>
#include <ESP8266WiFiMulti.h>
#include <WebSocketsClient.h> //  https://github.com/kakopappa/sinric/wiki/How-to-add-dependency-libraries
#include <ArduinoJson.h> // https://github.com/kakopappa/sinric/wiki/How-to-add-dependency-libraries (use the correct version)
#include <StreamString.h>
#include <DHTesp.h>
#include "Adafruit_MQTT.h"
#include "Adafruit_MQTT_Client.h"

ESP8266WiFiMulti WiFiMulti;
WebSocketsClient webSocket;
WiFiClient client;

#define MyApiKey "xxxxxxxxxxxxxxxx" // TODO: Change to your sinric API Key. Your API Key is displayed on sinric.com dashboard
#define MySSID "xxxxxxxxx" // TODO: Change to your Wifi network SSID
#define MyWifiPassword "xxxxxxxxxxxxxx" // TODO: Change to your Wifi network password
#define DeviceID "xxxxxxxxxxxxxxxxxxxxx" // ID Controller RGB
#define DeviceID2 "xxxxxxxxxxxxxxxxxxxxxxxxxx" // ID Sensore Temperatura

#define HEARTBEAT_INTERVAL 5*60*1000 // 5 Minutes 
#define TEMPRATURE_INTERVAL 1*60*1000 // 1 Minutes
#define IOT_INTERVAL 5*60*1000 // 5 Minutes
#define SERVER_URL "iot.sinric.com" //"iot.sinric.com"
#define SERVER_PORT 80 // 80

/************************* Information to Connect to Adafruit IO *********************************/
/*  You will need and account - Its FREE  */

#define AIO_SERVER      "io.adafruit.com"
#define AIO_SERVERPORT  1883                   // use 8883 for SSL
#define AIO_USERNAME    "xxxxxxx"
#define AIO_KEY         "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxx"

Adafruit_MQTT_Client mqtt(&client, AIO_SERVER, AIO_SERVERPORT, AIO_USERNAME, AIO_KEY);
Adafruit_MQTT_Publish FeedTemperatura = Adafruit_MQTT_Publish(&mqtt, AIO_USERNAME "/feeds/temperatura");
Adafruit_MQTT_Publish FeedUmidita = Adafruit_MQTT_Publish(&mqtt, AIO_USERNAME "/feeds/umidita");

uint64_t heartbeatTimestamp = 0;
uint64_t tempratureUpdateTimestamp = 0;
uint64_t TempUpdateTimestamp = 0;
uint64_t iotUpdateTimestamp = 0;
bool isConnected = false;
bool luce = false;
bool oldState = false;

void setPowerStateOnServer(String deviceId, String value);
void setSetTemperatureSettingOnServer(String deviceId, String value, String scale);
void readTempature();

float humidity;
float temperature;
float temperaturefh;

float hue = 0.0;
float sat = 0.0;
float bri = 0.0;
int brigh = 100;

DHTesp dht;

void MQTT_connect();

// deviceId is the ID assgined to your smart-home-device in sinric.com dashboard. Copy it from dashboard and paste it here

void turnOn(String deviceId) {
  if (deviceId == DeviceID) // Device ID of first device
  {
    Serial.print("Turn on device id: ");
    Serial.println(deviceId);
    luce = true;
  }
  else {
    Serial.print("Turn on for unknown device id: ");
    Serial.println(deviceId);
  }
}

void turnOff(String deviceId) {
  if (deviceId == DeviceID) // Device ID of first device
  {
    Serial.print("Turn off Device ID: ");
    Serial.println(deviceId);
    luce = false;
  }
  else {
    Serial.print("Turn off for unknown device id: ");
    Serial.println(deviceId);
  }
}

void webSocketEvent(WStype_t type, uint8_t * payload, size_t length) {
  switch (type) {
    case WStype_DISCONNECTED:
      isConnected = false;
      Serial.printf("[WSc] Webservice disconnected from sinric.com!\n");
      break;
    case WStype_CONNECTED: {
        isConnected = true;
        Serial.printf("[WSc] Service connected to sinric.com at url: %s\n", payload);
        Serial.printf("Waiting for commands from sinric.com ...\n");
      }
      break;
    case WStype_TEXT: {
        Serial.printf("[WSc] get text: %s\n", payload);
        // Example payloads

        // For Switch or Light device types
        // {"deviceId": xxxx, "action": "setPowerState", value: "ON"} // https://developer.amazon.com/docs/device-apis/alexa-powercontroller.html

        // For Light device type
        // Look at the light example in github

        DynamicJsonBuffer jsonBuffer;
        JsonObject& json = jsonBuffer.parseObject((char*)payload);
        String deviceId = json ["deviceId"];
        String action = json ["action"];
        if (deviceId == DeviceID)
        {
          if (action == "setPowerState") { // Switch or Light
            String value = json ["value"];
            if (value == "ON") {
              turnOn(deviceId);
            } else {
              turnOff(deviceId);
            }
          }
          else if (action == "SetColor") {
            // Alexa, set the device name to red
            // get text: {"deviceId":"xxxx","action":"SetColor","value":{"hue":0,"saturation":1,"brightness":1}}
            String Ahue = json ["value"]["hue"];
            String Asaturation = json ["value"]["saturation"];
            String Abrightness = json ["value"]["brightness"];

            hue = Ahue.toInt();
            sat = Asaturation.toDouble();
            bri = Abrightness.toDouble();

            Serial.println("[WSc] hue: " + Ahue);
            Serial.println("[WSc] saturation: " + Asaturation);
            Serial.println("[WSc] brightness: " + Abrightness);
            luce = true;
          }
          else if (action == "SetColorTemperature") {
            String Ahue = json ["value"];
            String Asaturation = "1";
            String Abrightness = "1";

            hue = Ahue.toDouble() / 55.0;
            sat = Asaturation.toDouble();
            bri = Abrightness.toDouble();
            luce = true;
          }
          else if (action == "SetBrightness") {
            String Abrightness = json ["value"];;

            brigh = Abrightness.toInt();
            luce = true;
          }
          else if (action == "test") {
            Serial.println("[WSc] received test command from sinric.com");
          }
        } else if (deviceId == DeviceID2) {
          Serial.println("[WSc] Ricevuto dato per termostato");
          readTempature();
          setSetTemperatureSettingOnServer(DeviceID2, temperature, "CELSIUS", temperature, humidity);
        }
        else {
          Serial.println("[WSc] Nope!");
        }
      }
      break;
    case WStype_BIN:
      Serial.printf("[WSc] get binary length: %u\n", length);
      break;
  }
}

void setup() {
  pinMode(ledR, OUTPUT);
  pinMode(ledG, OUTPUT);
  pinMode(ledB, OUTPUT);
  //pinMode(pinTemp, INPUT_PULLUP);

  dht.setup(pinTemp, DHTesp::DHT11); // Connect DHT sensor to GPIO

  Serial.begin(115200);

  WiFiMulti.addAP(MySSID, MyWifiPassword);
  Serial.println();
  Serial.print("Connecting to Wifi: ");
  Serial.println(MySSID);

  // Waiting for Wifi connect
  while (WiFiMulti.run() != WL_CONNECTED) {
    delay(500);
    Serial.print(".");
  }
  if (WiFiMulti.run() == WL_CONNECTED) {
    Serial.println("");
    Serial.print("WiFi connected. ");
    Serial.print("IP address: ");
    Serial.println(WiFi.localIP());
  }

  // server address, port and URL
  webSocket.begin(SERVER_URL, SERVER_PORT, "/");

  // event handler
  webSocket.onEvent(webSocketEvent);
  webSocket.setAuthorization("apikey", MyApiKey);

  // try again every 5000ms if connection has failed
  webSocket.setReconnectInterval(5000);
  // If you see 'class WebSocketsClient' has no member named 'setReconnectInterval' error update arduinoWebSockets
}

void loop() {
  MQTT_connect();
  webSocket.loop();
  //readTempature();

  uint64_t now = millis();

  if (isConnected) {
    // Send heartbeat in order to avoid disconnections during ISP resetting IPs over night. Thanks @MacSass
    if ((now - heartbeatTimestamp) > HEARTBEAT_INTERVAL) {
      heartbeatTimestamp = now;
      if (luce) {
        setPowerStateOnServer(DeviceID, "ON");
      } else {
        setPowerStateOnServer(DeviceID, "OFF");
      }
    }
    if (luce) {
      setLedColorHSV(hue, sat, bri);
    } else {
      digitalWrite(ledR, LOW);
      digitalWrite(ledG, LOW);
      digitalWrite(ledB, LOW);
    }
    if ((now - tempratureUpdateTimestamp) > TEMPRATURE_INTERVAL) {
      tempratureUpdateTimestamp = now;
      setSetTemperatureSettingOnServer(DeviceID2, temperature, "CELSIUS", temperature, humidity);
    }
    if (luce != oldState) {
      if (luce) {
        setPowerStateOnServer(DeviceID, "ON");
      } else {
        setPowerStateOnServer(DeviceID, "OFF");
      }
      oldState = luce;
    }
  }
  if ((now - iotUpdateTimestamp) > IOT_INTERVAL) {
    iotUpdateTimestamp = now;
    FeedTemperatura.publish(temperature);
    FeedUmidita.publish(humidity);
    Serial.println("[IOT] Pubblicato");
  }
  if ((now - TempUpdateTimestamp) > 15000) {
    TempUpdateTimestamp = now;
    readTempature();
  }
}

void MQTT_connect() {
  int8_t ret;

  // Stop if already connected.
  if (mqtt.connected()) {
    return;
  }

  Serial.print("Connecting to MQTT... ");

  uint8_t retries = 3;
  while ((ret = mqtt.connect()) != 0) { // connect will return 0 for connected
    Serial.println(mqtt.connectErrorString(ret));
    Serial.println("Retrying MQTT connection in 5 seconds...");
    mqtt.disconnect();
    delay(5000);  // wait 5 seconds
    retries--;
    if (retries == 0) {
      // basically die and wait for WDT to reset me
      while (1);
    }
  }
  Serial.println("MQTT Connected!");
}

void setLedColorHSV(int h, double s, double v) {
  //this is the algorithm to convert from RGB to HSV
  double r = 0;
  double g = 0;
  double b = 0;

  double hf = h / 60.0;

  int i = (int)floor(h / 60.0);
  double f = h / 60.0 - i;
  double pv = v * (1 - s);
  double qv = v * (1 - s * f);
  double tv = v * (1 - s * (1 - f));

  switch (i)
  {
    case 0: //rojo dominante
      r = v;
      g = tv;
      b = pv;
      break;
    case 1: //verde
      r = qv;
      g = v;
      b = pv;
      break;
    case 2:
      r = pv;
      g = v;
      b = tv;
      break;
    case 3: //azul
      r = pv;
      g = qv;
      b = v;
      break;
    case 4:
      r = tv;
      g = pv;
      b = v;
      break;
    case 5: //rojo
      r = v;
      g = pv;
      b = qv;
      break;
  }

  //set each component to a integer value between 0 and 255
  int red = constrain((int)255 * r, 0, 255);
  int green = constrain((int)255 * g, 0, 255);
  int blue = constrain((int)255 * b, 0, 255);

  setLedColor(red, green, blue);
}

//Sets the current color for the RGB LED
void setLedColor(int red, int green, int blue) {
  //Note that we are reducing 1/4 the intensity for the green and blue components because
  //  the red one is too dim on my LED. You may want to adjust that.
  analogWrite(ledR, (((red * brigh) / 100) * 4));
  analogWrite(ledG, (((green * brigh) / 100) * 4));
  analogWrite(ledB, (((blue * brigh) / 100) * 4));
}

void readTempature() {
  //delay(dht.getMinimumSamplingPeriod());

  humidity = dht.getHumidity();
  temperature = dht.getTemperature();
  temperaturefh = dht.toFahrenheit(temperature);

  Serial.print(dht.getStatusString());
  Serial.print("\t");
  Serial.print(humidity, 1);
  Serial.print("\t\t");
  Serial.print(temperature, 1);
  Serial.print("\t\t");
  Serial.print(dht.toFahrenheit(temperature), 1);
  Serial.print("\t\t");
  Serial.print(dht.computeHeatIndex(temperature, humidity, false), 1);
  Serial.print("\t\t");
  Serial.println(dht.computeHeatIndex(dht.toFahrenheit(temperature), humidity, true), 1);
}

// If you are going to use a push button to on/off the switch manually, use this function to update the status on the server
// so it will reflect on Alexa app.
// eg: setPowerStateOnServer("deviceid", "ON")

// Call ONLY If status changed. DO NOT CALL THIS IN loop() and overload the server.

void setPowerStateOnServer(String deviceId, String value) {
  DynamicJsonBuffer jsonBuffer;
  JsonObject& root = jsonBuffer.createObject();
  root["deviceId"] = deviceId;
  root["action"] = "setPowerState";
  root["value"] = value;
  StreamString databuf;
  root.printTo(databuf);

  webSocket.sendTXT(databuf);
}

// Call ONLY If status changed. DO NOT CALL THIS IN loop() and overload the server.

//eg: setSetTemperatureSettingOnServer("deviceid", 25.0, "CELSIUS" or "FAHRENHEIT", 23.0, 45.3)
// setPoint: Indicates the target temperature to set on the termostat.
void setSetTemperatureSettingOnServer(String deviceId, float setPoint, String scale, float ambientTemperature, float ambientHumidity) {
  DynamicJsonBuffer jsonBuffer;
  JsonObject& root = jsonBuffer.createObject();
  root["action"] = "SetTemperatureSetting";
  root["deviceId"] = deviceId;

  JsonObject& valueObj = root.createNestedObject("value");
  JsonObject& temperatureSetting = valueObj.createNestedObject("temperatureSetting");
  temperatureSetting["setPoint"] = setPoint;
  temperatureSetting["scale"] = scale;
  temperatureSetting["ambientTemperature"] = ambientTemperature;
  temperatureSetting["ambientHumidity"] = ambientHumidity;

  StreamString databuf;
  root.printTo(databuf);

  //Serial.println(databuf);
  webSocket.sendTXT(databuf);
}
// Call ONLY If status changed. DO NOT CALL THIS IN loop() and overload the server.

void setThermostatModeOnServer(String deviceId, String thermostatMode) {
  DynamicJsonBuffer jsonBuffer;
  JsonObject& root = jsonBuffer.createObject();
  root["deviceId"] = deviceId;
  root["action"] = "SetThermostatMode";
  root["value"] = thermostatMode;
  StreamString databuf;
  root.printTo(databuf);

  webSocket.sendTXT(databuf);
}
